@@ -20,22 +20,22 @@ import time
 from waitress import trigger
 from waitress.adjustments import Adjustments
 from waitress.channel import HTTPChannel
+from waitress.compat import IPPROTO_IPV6, IPV6_V6ONLY
 from waitress.task import ThreadedTaskDispatcher
 from waitress.utilities import cleanup_unix_socket
 
-from waitress.compat import (
-    IPPROTO_IPV6,
-    IPV6_V6ONLY,
-    )
 from . import wasyncore
+from .proxy_headers import proxy_headers_middleware
 
-def create_server(application,
-                  map=None,
-                  _start=True,      # test shim
-                  _sock=None,       # test shim
-                  _dispatcher=None, # test shim
-                  **kw              # adjustments
-                  ):
+
+def create_server(
+    application,
+    map=None,
+    _start=True,  # test shim
+    _sock=None,  # test shim
+    _dispatcher=None,  # test shim
+    **kw  # adjustments
+):
     """
     if __name__ == '__main__':
         server = create_server(app)
@@ -44,11 +44,11 @@ def create_server(application,
     if application is None:
         raise ValueError(
             'The "app" passed to ``create_server`` was ``None``.  You forgot '
-            'to return a WSGI app within your application.'
-            )
+            "to return a WSGI app within your application."
+        )
     adj = Adjustments(**kw)
 
-    if map is None: # pragma: nocover
+    if map is None:  # pragma: nocover
         map = {}
 
     dispatcher = _dispatcher
@@ -56,7 +56,7 @@ def create_server(application,
         dispatcher = ThreadedTaskDispatcher()
         dispatcher.set_thread_count(adj.threads)
 
-    if adj.unix_socket and hasattr(socket, 'AF_UNIX'):
+    if adj.unix_socket and hasattr(socket, "AF_UNIX"):
         sockinfo = (socket.AF_UNIX, socket.SOCK_STREAM, None, None)
         return UnixWSGIServer(
             application,
@@ -65,7 +65,8 @@ def create_server(application,
             _sock,
             dispatcher=dispatcher,
             adj=adj,
-            sockinfo=sockinfo)
+            sockinfo=sockinfo,
+        )
 
     effective_listen = []
     last_serv = None
@@ -81,8 +82,11 @@ def create_server(application,
                 _sock,
                 dispatcher=dispatcher,
                 adj=adj,
-                sockinfo=sockinfo)
-            effective_listen.append((last_serv.effective_host, last_serv.effective_port))
+                sockinfo=sockinfo,
+            )
+            effective_listen.append(
+                (last_serv.effective_host, last_serv.effective_port)
+            )
 
     for sock in adj.sockets:
         sockinfo = (sock.family, sock.type, sock.proto, sock.getsockname())
@@ -95,9 +99,12 @@ def create_server(application,
                 dispatcher=dispatcher,
                 adj=adj,
                 bind_socket=False,
-                sockinfo=sockinfo)
-            effective_listen.append((last_serv.effective_host, last_serv.effective_port))
-        elif hasattr(socket, 'AF_UNIX') and sock.family == socket.AF_UNIX:
+                sockinfo=sockinfo,
+            )
+            effective_listen.append(
+                (last_serv.effective_host, last_serv.effective_port)
+            )
+        elif hasattr(socket, "AF_UNIX") and sock.family == socket.AF_UNIX:
             last_serv = UnixWSGIServer(
                 application,
                 map,
@@ -106,8 +113,11 @@ def create_server(application,
                 dispatcher=dispatcher,
                 adj=adj,
                 bind_socket=False,
-                sockinfo=sockinfo)
-            effective_listen.append((last_serv.effective_host, last_serv.effective_port))
+                sockinfo=sockinfo,
+            )
+            effective_listen.append(
+                (last_serv.effective_host, last_serv.effective_port)
+            )
 
     # We are running a single server, so we can just return the last server,
     # saves us from having to create one more object
@@ -115,37 +125,41 @@ def create_server(application,
         # In this case we have no need to use a MultiSocketServer
         return last_serv
 
+    log_info = last_serv.log_info
     # Return a class that has a utility function to print out the sockets it's
     # listening on, and has a .run() function. All of the TcpWSGIServers
     # registered themselves in the map above.
-    return MultiSocketServer(map, adj, effective_listen, dispatcher)
+    return MultiSocketServer(map, adj, effective_listen, dispatcher, log_info)
 
 
 # This class is only ever used if we have multiple listen sockets. It allows
 # the serve() API to call .run() which starts the wasyncore loop, and catches
 # SystemExit/KeyboardInterrupt so that it can atempt to cleanly shut down.
-class MultiSocketServer(object):
-    asyncore = wasyncore # test shim
+class MultiSocketServer:
+    asyncore = wasyncore  # test shim
 
-    def __init__(self,
-                 map=None,
-                 adj=None,
-                 effective_listen=None,
-                 dispatcher=None,
-                 ):
+    def __init__(
+        self,
+        map=None,
+        adj=None,
+        effective_listen=None,
+        dispatcher=None,
+        log_info=None,
+    ):
         self.adj = adj
         self.map = map
         self.effective_listen = effective_listen
         self.task_dispatcher = dispatcher
+        self.log_info = log_info
 
-    def print_listen(self, format_str): # pragma: nocover
+    def print_listen(self, format_str):  # pragma: nocover
         for l in self.effective_listen:
             l = list(l)
 
-            if ':' in l[0]:
-                l[0] = '[{}]'.format(l[0])
+            if ":" in l[0]:
+                l[0] = "[{}]".format(l[0])
 
-            print(format_str.format(*l))
+            self.log_info(format_str.format(*l))
 
     def run(self):
         try:
@@ -162,26 +176,43 @@ class MultiSocketServer(object):
         wasyncore.close_all(self.map)
 
 
-class BaseWSGIServer(wasyncore.dispatcher, object):
+class BaseWSGIServer(wasyncore.dispatcher):
 
     channel_class = HTTPChannel
     next_channel_cleanup = 0
-    socketmod = socket # test shim
-    asyncore = wasyncore # test shim
+    socketmod = socket  # test shim
+    asyncore = wasyncore  # test shim
+    in_connection_overflow = False
 
-    def __init__(self,
-                 application,
-                 map=None,
-                 _start=True,      # test shim
-                 _sock=None,       # test shim
-                 dispatcher=None,  # dispatcher
-                 adj=None,         # adjustments
-                 sockinfo=None,    # opaque object
-                 bind_socket=True,
-                 **kw
-                 ):
+    def __init__(
+        self,
+        application,
+        map=None,
+        _start=True,  # test shim
+        _sock=None,  # test shim
+        dispatcher=None,  # dispatcher
+        adj=None,  # adjustments
+        sockinfo=None,  # opaque object
+        bind_socket=True,
+        **kw
+    ):
         if adj is None:
             adj = Adjustments(**kw)
+
+        if adj.trusted_proxy or adj.clear_untrusted_proxy_headers:
+            # wrap the application to deal with proxy headers
+            # we wrap it here because webtest subclasses the TcpWSGIServer
+            # directly and thus doesn't run any code that's in create_server
+            application = proxy_headers_middleware(
+                application,
+                trusted_proxy=adj.trusted_proxy,
+                trusted_proxy_count=adj.trusted_proxy_count,
+                trusted_proxy_headers=adj.trusted_proxy_headers,
+                clear_untrusted=adj.clear_untrusted_proxy_headers,
+                log_untrusted=adj.log_untrusted_proxy_headers,
+                logger=self.logger,
+            )
+
         if map is None:
             # use a nonglobal socket map by default to hopefully prevent
             # conflicts with apps and libs that use the wasyncore global socket
@@ -204,7 +235,7 @@ class BaseWSGIServer(wasyncore.dispatcher, object):
         self.asyncore.dispatcher.__init__(self, _sock, map=map)
         if _sock is None:
             self.create_socket(self.family, self.socktype)
-            if self.family == socket.AF_INET6: # pragma: nocover
+            if self.family == socket.AF_INET6:  # pragma: nocover
                 self.socket.setsockopt(IPPROTO_IPV6, IPV6_V6ONLY, 1)
 
         self.set_reuse_addr()
@@ -213,53 +244,20 @@ class BaseWSGIServer(wasyncore.dispatcher, object):
             self.bind_server_socket()
 
         self.effective_host, self.effective_port = self.getsockname()
-        self.server_name = self.get_server_name(self.effective_host)
+        self.server_name = adj.server_name
         self.active_channels = {}
         if _start:
             self.accept_connections()
 
     def bind_server_socket(self):
-        raise NotImplementedError # pragma: no cover
-
-    def get_server_name(self, ip):
-        """Given an IP or hostname, try to determine the server name."""
-
-        if not ip:
-            raise ValueError('Requires an IP to get the server name')
-
-        server_name = str(ip)
-
-        # If we are bound to all IP's, just return the current hostname, only
-        # fall-back to "localhost" if we fail to get the hostname
-        if server_name == '0.0.0.0' or server_name == '::':
-            try:
-                return str(self.socketmod.gethostname())
-            except (socket.error, UnicodeDecodeError):  # pragma: no cover
-                # We also deal with UnicodeDecodeError in case of Windows with
-                # non-ascii hostname
-                return 'localhost'
-
-        # Now let's try and convert the IP address to a proper hostname
-        try:
-            server_name = self.socketmod.gethostbyaddr(server_name)[0]
-        except (socket.error, UnicodeDecodeError):  # pragma: no cover
-            # We also deal with UnicodeDecodeError in case of Windows with
-            # non-ascii hostname
-            pass
-
-        # If it contains an IPv6 literal, make sure to surround it with
-        # brackets
-        if ':' in server_name and '[' not in server_name:
-            server_name = '[{}]'.format(server_name)
-
-        return server_name
+        raise NotImplementedError  # pragma: no cover
 
     def getsockname(self):
-        raise NotImplementedError # pragma: no cover
+        raise NotImplementedError  # pragma: no cover
 
     def accept_connections(self):
         self.accepting = True
-        self.socket.listen(self.adj.backlog) # Get around asyncore NT limit
+        self.socket.listen(self.adj.backlog)  # Get around asyncore NT limit
 
     def add_task(self, task):
         self.task_dispatcher.add_task(task)
@@ -269,7 +267,28 @@ class BaseWSGIServer(wasyncore.dispatcher, object):
         if now >= self.next_channel_cleanup:
             self.next_channel_cleanup = now + self.adj.cleanup_interval
             self.maintenance(now)
-        return (self.accepting and len(self._map) < self.adj.connection_limit)
+
+        if self.accepting:
+            if (
+                not self.in_connection_overflow
+                and len(self._map) >= self.adj.connection_limit
+            ):
+                self.in_connection_overflow = True
+                self.logger.warning(
+                    "total open connections reached the connection limit, "
+                    "no longer accepting new connections"
+                )
+            elif (
+                self.in_connection_overflow
+                and len(self._map) < self.adj.connection_limit
+            ):
+                self.in_connection_overflow = False
+                self.logger.info(
+                    "total open connections dropped below the connection limit, "
+                    "listening again"
+                )
+            return not self.in_connection_overflow
+        return False
 
     def writable(self):
         return False
@@ -286,14 +305,13 @@ class BaseWSGIServer(wasyncore.dispatcher, object):
             if v is None:
                 return
             conn, addr = v
-        except socket.error:
+        except OSError:
             # Linux: On rare occasions we get a bogus socket back from
             # accept.  socketmodule.c:makesockaddr complains that the
             # address family is unknown.  We don't want the whole server
             # to shut down because of this.
             if self.adj.log_socket_errors:
-                self.logger.warning('server accept() threw an exception',
-                                    exc_info=True)
+                self.logger.warning("server accept() threw an exception", exc_info=True)
             return
         self.set_socket_options(conn)
         addr = self.fix_addr(addr)
@@ -329,8 +347,8 @@ class BaseWSGIServer(wasyncore.dispatcher, object):
             if (not channel.requests) and channel.last_activity < cutoff:
                 channel.will_close = True
 
-    def print_listen(self, format_str): # pragma: nocover
-        print(format_str.format(self.effective_host, self.effective_port))
+    def print_listen(self, format_str):  # pragma: no cover
+        self.log_info(format_str.format(self.effective_host, self.effective_port))
 
     def close(self):
         self.trigger.close()
@@ -338,50 +356,40 @@ class BaseWSGIServer(wasyncore.dispatcher, object):
 
 
 class TcpWSGIServer(BaseWSGIServer):
-
     def bind_server_socket(self):
         (_, _, _, sockaddr) = self.sockinfo
         self.bind(sockaddr)
 
     def getsockname(self):
-        try:
-            return self.socketmod.getnameinfo(
-                self.socket.getsockname(),
-                self.socketmod.NI_NUMERICSERV
-            )
-        except: # pragma: no cover
-            # This only happens on Linux because a DNS issue is considered a
-            # temporary failure that will raise (even when NI_NAMEREQD is not
-            # set). Instead we try again, but this time we just ask for the
-            # numerichost and the numericserv (port) and return those. It is
-            # better than nothing.
-            return self.socketmod.getnameinfo(
-                self.socket.getsockname(),
-                self.socketmod.NI_NUMERICHOST | self.socketmod.NI_NUMERICSERV
-            )
+        # Return the IP address, port as numeric
+        return self.socketmod.getnameinfo(
+            self.socket.getsockname(),
+            self.socketmod.NI_NUMERICHOST | self.socketmod.NI_NUMERICSERV,
+        )
 
     def set_socket_options(self, conn):
         for (level, optname, value) in self.adj.socket_options:
             conn.setsockopt(level, optname, value)
 
 
-if hasattr(socket, 'AF_UNIX'):
+if hasattr(socket, "AF_UNIX"):
 
     class UnixWSGIServer(BaseWSGIServer):
-
-        def __init__(self,
-                     application,
-                     map=None,
-                     _start=True,      # test shim
-                     _sock=None,       # test shim
-                     dispatcher=None,  # dispatcher
-                     adj=None,         # adjustments
-                     sockinfo=None,    # opaque object
-                     **kw):
+        def __init__(
+            self,
+            application,
+            map=None,
+            _start=True,  # test shim
+            _sock=None,  # test shim
+            dispatcher=None,  # dispatcher
+            adj=None,  # adjustments
+            sockinfo=None,  # opaque object
+            **kw
+        ):
             if sockinfo is None:
                 sockinfo = (socket.AF_UNIX, socket.SOCK_STREAM, None, None)
 
-            super(UnixWSGIServer, self).__init__(
+            super().__init__(
                 application,
                 map=map,
                 _start=_start,
@@ -389,7 +397,8 @@ if hasattr(socket, 'AF_UNIX'):
                 dispatcher=dispatcher,
                 adj=adj,
                 sockinfo=sockinfo,
-                **kw)
+                **kw,
+            )
 
         def bind_server_socket(self):
             cleanup_unix_socket(self.adj.unix_socket)
@@ -398,13 +407,11 @@ if hasattr(socket, 'AF_UNIX'):
                 os.chmod(self.adj.unix_socket, self.adj.unix_socket_perms)
 
         def getsockname(self):
-            return ('unix', self.socket.getsockname())
+            return ("unix", self.socket.getsockname())
 
         def fix_addr(self, addr):
-            return ('localhost', None)
+            return ("localhost", None)
 
-        def get_server_name(self, ip):
-            return 'localhost'
 
 # Compatibility alias.
 WSGIServer = TcpWSGIServer
